import enum
from .. import create_commander, var, CommandResult
from ..transformation import to_bool, to_enum


class OutputFormat(enum.Enum):
    HUMAN = 'human'
    JSON = 'json'


command = create_commander('set', description='Change properties of the console interface.')


@command('set', is_abstract=True, inject=['console'])
def set_state():
    pass


@command('echo', var('state', transform=to_bool, choices=('on', 'off')), parent=set_state)
def set_echo(console, state):
    """ enable/disable any output of prompts or typed text """
    console.set_echo(state)
    yield CommandResult('Echo is now {}'.format('enabled' if state else 'disabled'))


@command('output-format', var('format', choices=OutputFormat), parent=set_state)
@command.transform(format=to_enum(OutputFormat))
def set_output_format(console, format: OutputFormat):
    """ switch the response output format """
    console.select_output_format(format)
    yield CommandResult('Output format set to "{}"'.format(format.value))



import collections
import enum
from functools import update_wrapper, wraps
from inspect import signature
from itertools import groupby
from typing import List, Sequence, Callable, Any, Optional, Iterator, Set
from blinker import signal
from .validation import validate, validate_value, one_of, ValidationError
from .transformation import BuildTimeTransformContext, transform, transform_value
from .utils import tokenize_args, join_generator_string, TokenMismatch, TaggedString, noop
from .error import CommandError, InjectionError


class _Undefined:
    pass


class CommandResult:
    def __init__(self, message, data=None, success=True, status=None) -> None:
        self.message = message
        self.data = data
        self.success = success
        self.status = status if status is not None else 0

    def __str__(self):
        return self.message


def _is_enum(choices):
    try:
        return issubclass(choices, enum.Enum)
    except TypeError:
        # "choices" is not a class (and thus not an enum)
        return False


class _Token:
    is_variable = False
    is_placeholder = False

    def __init__(self, value=_Undefined, name=None, is_optional=False,
                 greedy=False, transform=None, choices=None, description=None,
                 aliases=None):
        """ specify possible value of a command string token

        @param choices: may be None, an enum or an iterable
        """
        # TODO: greedy should accept integers
        self.value = value
        self.name = name
        self.is_optional = is_optional
        self.greedy = greedy
        self.transform = transform
        self.description = description
        self.aliases = frozenset(aliases or [])
        if choices is None:
            self.choices = set()
        elif _is_enum(choices):
            self.choices = {item.value for item in choices}
        else:
            self.choices = set(choices)

    def matches(self, value):
        if self.value is not _Undefined:
            return self.value == value or value in self.aliases
        if self.name and self.transform:
            # we trigger the token validation here to handle args that
            # have been misspelled and to provide command suggestions
            getattr(self.transform, 'validate', noop)(value)
        if self.name and self.choices:
            # if choices were defined we must validate that the provided
            # value actually is one of those choices
            validate_value(one_of(self.choices), value)
        return True

    def get_label(self, with_error=None):
        def format_greedy(name):
            return '[%s [...]]' % name

        def format_optional(name):
            name = TaggedString.error(name) if with_error else TaggedString.optional(name)
            return '[%s]' % name

        def get_name(token):
            if token.name:
                return token.name.upper()
            if token.value:
                return token.value
            else:
                raise ValueError('cannot derive name from token')

        token = get_name(self)

        if self.greedy:
            token = format_greedy(token)
        if self.is_optional:
            token = format_optional(token)
        if self.name and not self.is_optional:
            token = TaggedString.var(token)

        return token

    def __str__(self):
        return self.get_label()


class _Placeholder(_Token):
    is_placeholder = True


class _Variable(_Token):
    is_variable = True


class _Group:
    def __init__(self, name, description=None):
        self.name = name
        self.description = description


def var(name, value=_Undefined, is_optional=False, transform=None, greedy=None,
        choices=None, description=None):
    return _Variable(value, name, is_optional, greedy, transform, choices, description)


def optional(value):
    return _Placeholder(value, is_optional=True)


def keyword(value, aliases=None):
    return _Token(value, aliases=aliases)


def group(name, description=None):
    return _Group(name, description)


def _find_token(tokens, index):
    if len(tokens) > index:
        return tokens[index]
    elif tokens[-1].greedy:
        return tokens[-1]
    else:
        raise IndexError('invalid token index')


class _MatchType(enum.Enum):
    NONE = None
    INVALID = 'invalid'
    PARTIAL = 'partial'
    EXACT = 'exact'

    def __str__(self):
        if self.value is _MatchType.NONE:
            raise NotImplementedError('please override the match type')
        return {
            _MatchType.INVALID: 'Invalid commands',
            _MatchType.PARTIAL: 'Partial matching commands',
            _MatchType.EXACT: 'Selected command'
        }[self]

    def __lt__(self, other):
        return self.value < other.value


class _CommandMatch:
    type = _MatchType.NONE

    def __init__(self, command):
        self.command = command


class _InvalidCommandMatch(_CommandMatch):
    type = _MatchType.INVALID


class _PartialCommandMatch(_CommandMatch):
    type = _MatchType.PARTIAL

    def __init__(self, command, token=None, token_mismatches: Sequence[TokenMismatch] = tuple()):
        super().__init__(command)
        self.token = token
        self.token_mismatches = token_mismatches


class _ExactCommandMatch(_CommandMatch):
    type = _MatchType.EXACT


class AmbiguousCommandError(RuntimeError):
    def __init__(self, matches: Sequence[_ExactCommandMatch], *args: Any) -> None:
        self.matches = matches
        super().__init__(*args)


class _CommandMatches:
    # todo should probably be a UserList
    def __init__(self, matches: Sequence[_CommandMatch]):
        self.matches = matches

    def describe(self, args, discard_invalid=True) -> str:
        @join_generator_string()
        def _render() -> str:
            yield 'provided args:'
            yield '\t%s' % ' '.join(args)
            yield ''
            exclude = (_MatchType.INVALID, ) if discard_invalid else None
            match_groups = self._filter_matches(exclude=exclude, group=True)
            for match_type, matches in match_groups:
                yield '%s' % str(match_type)
                for match in matches:
                    token_mismatches = getattr(match, 'token_mismatches', None)
                    yield '\t%s' % match.command.get_label(with_errors=token_mismatches)
        return _render()

    def _filter_matches(self, include=None, exclude=None, group=False) -> List[_CommandMatch]:
        def _pluck_type(m):
            return m.type

        def _filter(match):
            if include and _pluck_type(match) in include:
                return True
            if exclude and _pluck_type(match) in exclude:
                return False

        matches = sorted(filter(_filter, self.matches), key=_pluck_type)
        return groupby(matches, key=_pluck_type) if group else matches

    def find_suggestable_matches(self, resolve, args):
        matches = self._filter_matches(include=(_MatchType.EXACT, _MatchType.PARTIAL))
        if matches:
            return matches
        elif not matches and len(args) >= 2:
            new_args = args[:-1]
            return resolve(new_args).find_suggestable_matches(resolve, new_args)
        else:
            return []

    def suggestion(self, resolve, args) -> str:
        @join_generator_string()
        def _render() -> str:
            matches = self.find_suggestable_matches(resolve, args)
            if len(matches) == 1:
                # a single match has been found for the args. this is likely
                # a command that was used almost correctly
                match = matches[0]
                if match.type is _MatchType.EXACT:
                    # possible matches are resolved by skipping arguments from end
                    # to start. if a command had too many arguments we may find an
                    # exact match at some point. this is the case here
                    yield 'You have provided too many arguments for this command.'
                    yield 'Usage:\n\t' + match.command.get_label()
                else:
                    # if it’s not an exact match it’s a mismatch, because one of
                    # the tokens was rejected (likely during validation).
                    # display errors for each mismatched token.
                    if len(match.token_mismatches) > 0:
                        for mismatch in match.token_mismatches:
                            yield '{name}: {message}'.format(
                                name=mismatch.token.get_label(), message=str(mismatch.exception))
                    else:
                        yield 'You have not provided sufficient arguments for this command.'
                        yield 'Usage:\n\t' + match.command.get_label()
            else:
                yield 'Could not find the command for "%s"' % ' '.join(args)
                if matches:
                    if len(matches) == 1:
                        yield 'Did you mean:'
                    else:
                        yield 'Did you mean one of:'
                    for match in matches:
                        yield '\t%s' % match.command.get_label()
        return _render()

    @property
    def exact_match(self) -> Optional[_ExactCommandMatch]:
        matches = [match for match in self.matches if isinstance(match, _ExactCommandMatch)]
        if len(matches) > 1:
            raise AmbiguousCommandError(matches)
        elif len(matches) == 1:
            return matches[0]
        else:
            return None


class _Command:
    def __init__(self, func, tokens: List[_Token], parent, is_abstract, group, inject,
                 is_disabled):
        self.func = func
        self.tokens = tokens
        self.parent = parent
        self.is_abstract = is_abstract
        self._group = group
        self.inject = inject
        self._is_disabled = is_disabled

    def __call__(self, args, require):
        func_args = _build_args(self.all_tokens, args)
        inject_args = _build_inject_args(self.func, self.all_injections, require)
        try:
            return self.func(**func_args, **inject_args)
        except ValidationError as exc:
            if exc.field:
                for token in self.all_tokens:
                    if token.name == exc.field:
                        exc.field = token
                        break
            raise exc

    def __str__(self):
        return ' '.join([str(token) for token in self.all_tokens])

    def __getattr__(self, name: str) -> Any:
        # this object acts as a wrapper for the underlying function
        # so it’s a good idea to look for any missing attributes on
        # the function itself
        return getattr(self.func, name)

    def match(self, args):
        if not self.is_executable:
            return _InvalidCommandMatch(self)

        tokens = self.all_tokens
        invalid_tokens = []

        for index, arg in enumerate(args):
            # first try to find a matching token for this arg
            try:
                token = _find_token(tokens, index)
            except IndexError:
                # we have more args than this command has tokens
                # so this is clearly invalid. greedy tokens have
                # already been taken into account
                return _InvalidCommandMatch(self)

            try:
                # see if the token matches the arg
                if not token.matches(arg):
                    # if the token does not match we can assume
                    # that this command is not valid
                    return _InvalidCommandMatch(self)
            except ValidationError as exc:
                # if the token matching raises a ValidationError
                # this command might still be valid, but a variable
                # was provided in an incorrect format
                if token.is_variable:
                    invalid_tokens.append(TokenMismatch(token, exc, arg))
                else:
                    return _InvalidCommandMatch(self)

        # we made sure that the provided args match the defined tokens
        # but we may have more tokens. if one of the left over tokens
        # is not optional this command is a partial match
        for token in tokens[len(args):]:
            if not token.is_optional:
                return _PartialCommandMatch(self, token, invalid_tokens)

        if invalid_tokens:
            return _PartialCommandMatch(self, None, invalid_tokens)
        else:
            return _ExactCommandMatch(self)

    def mount(self, parent):
        self.parent = parent

    def get_label(self, with_errors: Sequence[TokenMismatch] = None):
        def find_error(token):
            if with_errors is None:
                return None
            else:
                for mismatch in with_errors:
                    if token == mismatch.token:
                        return mismatch.exception
            return None

        return ' '.join(
            token.get_label(with_error=find_error(token))
            for token in self.all_tokens
        )

    @property
    def is_disabled(self):
        return self._is_disabled(self) if callable(self._is_disabled) \
            else bool(self._is_disabled)

    @property
    def is_executable(self):
        return not self.is_abstract and not self.is_disabled

    @property
    def group(self):
        if self._group:
            return self._group
        if self.parent:
            return self.parent.group
        else:
            return None

    @property
    def all_tokens(self):
        if self.parent:
            return self.parent.all_tokens + self.tokens
        else:
            return self.tokens

    @property
    def all_injections(self):
        if self.parent:
            return self.parent.all_injections + self.inject
        else:
            return self.inject


def _build_inject_args(func, injections, require):
    result_args = {}
    for injection in injections:
        key, value = injection.resolve(func, require)
        result_args[key] = value
    return result_args


def _build_args(tokens, args: Sequence):
    args = list(args)
    result_args = {}
    for token in tokens:
        # we might still have optional tokens but no more args so we
        # stop the processing once we depleted our args
        if len(args) == 0:
            break

        if token.name:
            def _transform(value):
                if BuildTimeTransformContext.takes_context(token.transform):
                    value = BuildTimeTransformContext(value, dict(result_args))
                return transform_value(token.transform, value)
            arg = args[0:] if token.greedy else args.pop(0)
            if token.transform and isinstance(arg, collections.Iterable) and \
                    not isinstance(arg, str):
                result = list(map(lambda x: _transform(x), arg))
            elif token.transform:
                result = _transform(arg)
            else:
                result = arg
            result_args[token.name] = result
        else:
            args.pop(0)
    return result_args


class _Injection:
    def __init__(self, key, inject_as, collect, default) -> None:
        self.key = key
        self.inject_as = inject_as or key
        self.collect = collect
        self.default = default

    def _get_default(self, func):
        sig = signature(func)
        if self.key in sig.parameters:
            parameter = sig.parameters[self.key]
            if parameter.default is not parameter.empty:
                return parameter.default
        return self.default

    def resolve(self, func, require):
        generator = require(self.key)
        try:
            value = list(generator) if self.collect else next(generator)
        except StopIteration as exc:
            default = self._get_default(func)
            if default is _Undefined:
                raise InjectionError(
                    'Dependency "{source}" injected as "{name}" was not provided at runtime. '
                    'Did you forget to call `command.provide("{source}", foo)` '
                    'at some point or to set a default value?'.format(
                        name=self.inject_as, source=self.key)
                ) from exc
            else:
                value = default
        return self.inject_as, value


def require(key=None, inject_as=None, collect=False, default=_Undefined):
    return _Injection(key, inject_as, collect, default)


def create_commander(name, description=None):
    commands = []
    context = {}
    static_context = {
        '__commands': commands
    }

    require_signal = signal('require')

    def _collect_commands():
        return filter_public_commands(commands)

    def _resolve_commands(args) -> _CommandMatches:
        return _CommandMatches([cmd.match(args) for cmd in _collect_commands()])

    def _require(key) -> Iterator[Any]:
        if key in static_context:
            yield static_context[key]
        if key in context:
            value, is_getter = context[key]
            yield value() if is_getter else value
        for response in require_signal.send(key):
            receiver, value = response
            if value is not None and receiver is not _require:
                yield from value

    class _provide:
        def __init__(self, key, obj, is_getter=False) -> None:
            self.key = key
            self.obj = obj
            context[key] = (obj, is_getter)

        def __enter__(self):
            pass

        def __exit__(self, *exc):
            if self.key not in context:
                return
            if self.obj is context[self.key][0]:
                context.pop(self.key)

    class Commander:
        provide = _provide

        def __init__(self, name):
            self.name = name
            self.__doc__ = """%s""" % (description or '')
            self.__class__.__name__ = name

        @staticmethod
        def validate(**kwargs):
            return validate(**kwargs)

        @staticmethod
        def transform(**kwargs):
            return transform(**kwargs)

        @staticmethod
        @tokenize_args()
        def dispatch(args, simulate=False) -> Iterator[CommandResult]:
            resolved_commands = _resolve_commands(args)

            if simulate:
                yield CommandResult(message=resolved_commands.describe(args))
            elif resolved_commands.exact_match:
                try:
                    yield from resolved_commands.exact_match.command(args, require=_require)
                except ValidationError as exc:
                    raise CommandError(
                        'Invalid argument{}: {}'.format(
                            ' for field ' + str(exc.field) if exc.field else '', str(exc)),
                        code='INVALID_ARGUMENT_FORMAT') from exc
            else:
                raise CommandError(resolved_commands.suggestion(_resolve_commands, args),
                                   code='INVALID_COMMAND')

        @classmethod
        def fire(cls, args):
            return list(cls.dispatch(args))

        @staticmethod
        def inject(*arg_requires, **kwarg_requires) -> Callable:
            arg_requires = _normalize_injections(arg_requires)
            kwarg_requires = _normalize_injections(kwarg_requires)

            def decorator(func):
                @wraps(func)
                def wrapper(*args, **kwargs):
                    arg_injections = _build_inject_args(func, arg_requires, _require)
                    kwarg_injections = _build_inject_args(func, kwarg_requires, _require)
                    injections = _merge_dicts(arg_injections, kwarg_injections)
                    fargs, fkwargs = _fit_args(func, args, _merge_dicts(injections, kwargs))
                    return func(*fargs, **fkwargs)
                return wrapper
            return decorator

        def __call__(self, *tokens, parent=None, is_abstract=False, group=None, inject=None,
                     is_disabled=False) -> Callable:
            tokens = _normalize_tokens(tokens)
            injections = _normalize_injections(inject)

            def decorator(func):
                func_command = _Command(func, tokens, parent, is_abstract,
                                        group, injections, is_disabled)
                update_wrapper(func_command, func)
                func_command.__commander__ = self
                commands.append(func_command)
                return func_command
            return decorator

        def compose(self, *commanders) -> 'Commander':
            for commander in commanders:
                commander.extend(commands, _require)
            return self

        def extend(self, your_commands, your_require) -> 'Commander':
            your_commands.extend(commands)
            require_signal.connect(your_require)
            return self

        def __str__(self):
            return name

    return Commander(name)


def _merge_dicts(*dicts):
    result = {}
    for d in dicts:
        result.update(d)
    return result


def _fit_args(func, args, kwargs):
    args = list(args)
    func_args = func.__code__.co_varnames[:func.__code__.co_argcount]
    for index, arg in enumerate(func_args):
        if arg in kwargs:
            args.insert(index, kwargs.pop(arg))
    return args, kwargs


def _normalize_tokens(tokens) -> List[_Token]:
    result = []
    for token in tokens:
        if not isinstance(token, _Token):
            token = _Token(value=token)
        result.append(token)

    return result


def _normalize_injections(injections) -> List[_Injection]:
    result = []

    if injections is None:
        return result

    def normalize_require(key):
        if key.endswith('[]'):
            return key[:-2], True
        else:
            return key, False

    if isinstance(injections, dict):
        for inject_as, value in injections.items():
            if isinstance(value, _Injection):
                value.key = value.inject_as = inject_as
                result.append(require(value.key, inject_as, collect=value.collect))
            else:
                value, collect = normalize_require(value)
                result.append(require(value, inject_as, collect))
    else:
        for injection in injections:
            if not isinstance(injection, _Injection):
                try:
                    key, inject_as = injection.split(':')
                    key, collect = normalize_require(key)
                except ValueError:
                    key, collect = normalize_require(injection)
                    inject_as = key
                injection = require(key, inject_as, collect)
            result.append(injection)
    return result


def filter_public_commands(commands) -> List[_Command]:
    return [command for command in commands if command.is_executable]


def filter_root_commands(commands) -> Set[_Command]:
    return set([command for command in commands if command.parent is None])



import curses
import json
# merely importing readline enable command history via "input"
# noinspection PyUnresolvedReferences
import readline  # noqa: F401
import sys
import termios

import blessings

from ..command.set import OutputFormat
from ..utils import strip_tags, render_tags


class Console:

    def __init__(self, output, prompt='> ', output_format=OutputFormat.HUMAN):
        """ initialize a console environment

        @param output: output file (e.g. sys.stdin) or callable returning such a file
        @param prompt: the prompt string to be output in front of every line (with "echo" enabled)
        @param output_format: the initial output format to be used by the interface
        """
        self._given_output = output
        self._prompt = prompt
        self._show_echo = True
        self.terminal = None  # type: blessings.Terminal
        self.linesep = "\n\r"
        self.select_output_format(output_format)
        self._last_status = 0

    @property
    def output(self):
        """ allow the use of a callable as an "output" source

        This ability is relevant for files, that may be replaced during runtime (e.g. disconnected
        USB gadget host).
        """
        if callable(self._given_output):
            return self._given_output()
        else:
            return self._given_output

    def __enter__(self):
        """ store the original settings of the terminal """
        try:
            self._original_console_attributes = termios.tcgetattr(self.output)
        except termios.error:
            # it may fail for piped output
            self._original_console_attributes = None
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """ restore the original settings of the terminal """
        if self._original_console_attributes is not None:
            tcsetattr_flags = termios.TCSAFLUSH
            if hasattr(termios, 'TCSASOFT'):
                tcsetattr_flags |= termios.TCSASOFT
            termios.tcsetattr(self.output, tcsetattr_flags, self._original_console_attributes)

    def select_output_format(self, output_format):
        self._output_format = output_format

    def set_echo(self, enabled):
        if self._original_console_attributes is not None:
            attributes = termios.tcgetattr(self.output)
            if enabled:
                attributes[3] |= termios.ECHO
            else:
                attributes[3] &= ~termios.ECHO
            tcsetattr_flags = termios.TCSAFLUSH
            if hasattr(termios, 'TCSASOFT'):
                tcsetattr_flags |= termios.TCSASOFT
            termios.tcsetattr(self.output, tcsetattr_flags, attributes)
        self._show_echo = enabled

    def configure_auto(self, force_disable_style=False):
        """ disable echo for non-interactive sessions """
        self.set_echo(sys.stdin.isatty())
        if self.output.isatty() and not force_disable_style:
            try:
                self.terminal = blessings.Terminal(stream=self.output)
            except curses.error:
                self.terminal = None
        else:
            self.terminal = None

    def linefeed(self):
        if self._show_echo:
            print(file=self.output, end=self.linesep)

    def _format_output(self, s):
        if self.terminal and self.terminal.number_of_colors:
            return render_tags(s, self.terminal)
        else:
            return strip_tags(s)

    def send_data(self, result):
        status = getattr(result, 'status', self._last_status)
        error_code = getattr(result, 'code', None)
        if self._output_format not in OutputFormat:
            raise NotImplementedError('Unknown output format selected: {}'
                                      .format(self._output_format))

        if self._output_format is OutputFormat.HUMAN:
            content = self._format_output(str(result))
        else:
            if self._output_format is OutputFormat.JSON:
                serializer = json.dumps
            else:
                raise NotImplementedError('no serializer for output formatted: {}'
                                          .format(self._output_format))
            content = serializer({
                'data': result.data,
                'status': status,
                'code': error_code
            })

        self._last_status = status
        end = ('\x20' if result.success else '\x07') + '\0' + self.linesep
        print(content, file=self.output, end=end)

    def get_prompt(self):
        if self._output_format == OutputFormat.HUMAN and self._show_echo:
            prompt = self._prompt() if callable(self._prompt) else self._prompt
        else:
            prompt = None
        if prompt and self.terminal:
            def _format(formatter, text, placeholder='___PLACEHOLDER___'):
                format_begin, format_end = formatter(placeholder).split(placeholder)
                # readline will fail to correctly determine the length of the prompt
                # if the prompt uses color codes as is. \001 and \002 are readline
                # escape sequences that will ignore any output in between
                # see: https://stackoverflow.com/a/9468954
                return '\001{format_begin}\002{text}\001{format_end}\002'.format(
                    format_begin=format_begin, text=text, format_end=format_end
                )
            if self._last_status > 0:
                prompt = _format(self.terminal.red, prompt)
            else:
                prompt = _format(self.terminal.blue, prompt)
        return prompt



from collections import namedtuple, Iterable, UserString
from functools import wraps
import importlib
import logging
import os
import inspect
import time
import math
import re
import shlex
from typing import Sequence, Callable
import blessings
from .events import StopProcessingEvent


PATH_ATTRIBUTE = re.compile(r'^(?P<attr>[a-zA-Z_][a-zA-Z_0-9]*)$')
PATH_DICT = re.compile(r'^(?P<attr>[a-zA-Z_][a-zA-Z_0-9]*)\[[\"\'](?P<key>.+)[\"\']\]$')
PATH_INDEX = re.compile(r'^(?P<attr>[a-zA-Z_][a-zA-Z_0-9]*)\[(?P<index>\d+)\]$')
PATH_CALL = re.compile(r'^(?P<attr>[a-zA-Z_][a-zA-Z_0-9]*)\(\)$')
TAG_REGEX = re.compile(r'(?:<(?P<tag>[a-z]+)>)'
                       r'(?P<content>(?:<(?!/)|[^<])+)'
                       r'(?:</(?P<closing_tag>[a-z]+)>)')
TokenMismatch = namedtuple('TokenMismatch', ['token', 'exception', 'value'])


def tokenize_args(comment_characters=('#',)):
    def decorator(func):
        @wraps(func)
        def wrapper(command, *args, **kwargs):
            if isinstance(command, str):
                command = command.strip()
                if any(map(lambda c: command.startswith(c), comment_characters)):
                    raise StopProcessingEvent()
                command = shlex.split(command)
            return func(command, *args, **kwargs)
        return wrapper
    return decorator


def is_or_extends(item, cls):
    if isinstance(item, cls):
        return True
    item_type = type(item)
    return issubclass(item_type, cls)


def join_generator_string(glue: str = os.linesep, formatter: Callable = None):
    def _format(value):
        if formatter is None:
            return str(value)
        return formatter(str(value))

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            return glue.join(map(_format, result)) if isinstance(result, Iterable) \
                else _format(result)
        return wrapper
    return decorator


def columns(separator='\t', join_char=None):
    join_char = join_char or separator

    def get_column_width(rows, index):
        return [len(row[index]) for row in rows]

    def calculate_column_widths(lines):
        column_lengths = []
        for index in range(0, len(lines[0])):
            column_lengths.append(max(get_column_width(lines, index)))
        return column_lengths

    def fit_row(row: Sequence[str], column_widths: Sequence[int]):
        for index, column in enumerate(row):
            yield column.ljust(column_widths[index])

    def decorator(func):
        @join_generator_string(formatter=str.rstrip)
        @wraps(func)
        def wrapper(*args, **kwargs):
            rows = [line.split(separator) for line in func(*args, *kwargs).split(os.linesep)]
            column_widths = calculate_column_widths(rows)
            for row in rows:
                yield join_char.join(fit_row(row, column_widths))
        return wrapper
    return decorator


class _Stop(Exception):
    pass


def throttle(frequency=None, limit=None):
    """
    throttle the frequency and limit the maximum number of calls of a function
    :param frequency: frequency of calls in hertz, defaults to infinity
    :param limit: total number of calls, defaults to infinity, must be > 0
    :raises _Stop: if limit has been reached
    :return: callable
    """
    frequency = math.inf if frequency is None else frequency
    calls_per_second = 1 / frequency
    # because variable scopes in python are pretty much useless for closures
    # we’re using single-element lists here
    last_call = [None]
    call_count = [0]

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            current_time = time.time()
            if last_call[0] is None or last_call[0] + calls_per_second < current_time:
                last_call[0] = current_time
                call_count[0] += 1
                result = func(*args, **kwargs)
                if inspect.isgenerator(result):
                    yield from result
                else:
                    yield result

            if call_count[0] == limit:
                raise throttle.Stop()
        return wrapper
    return decorator


throttle.Stop = _Stop


def enum_value_list(enum):
    return [item.value for item in tuple(enum)]


def _tagged_string(tag):
    @classmethod
    def func(cls, s):
        return str(cls(s, tag))
    func.__name__ = tag
    return func


class TaggedString(UserString):
    def __init__(self, seq, tag):
        super().__init__(seq)
        self.tag = tag

    def __str__(self):
        return '<{tag}>{token}</{tag}>'.format(token=self.data, tag=self.tag)

    @classmethod
    def __getattr__(cls, item):
        return lambda s: _tagged_string(item)(cls, s)

    optional = _tagged_string('optional')
    var = _tagged_string('var')
    error = _tagged_string('error')
    help = _tagged_string('help')
    label = _tagged_string('label')
    header = _tagged_string('header')


def strip_tags(s):
    # we don’t process html here but our own simple one-line strings
    # if you were to use this for html you are a bad bad dog
    return re.sub(r'<[^<]+?>', '', s)


def render_tags(s, terminal: blessings.Terminal):
    def _replace_tag(match):
        tag, content, closing_tag = match.groups()
        if tag != closing_tag:
            raise ValueError('tag mismatch in content')
        replaced_value = {
            'optional': terminal.dim,
            'var': terminal.bold_white,
            'error': terminal.red,
            'label': terminal.bold,
            'help': terminal.italic_dim,
            'header': lambda s: terminal.bold(s.upper()),
        }[tag](content)
        return replaced_value

    # normally re.sub would replace all occurrences in our
    # text but tags may be nested so we better be safe and
    # replace tags as long as we find some
    while re.search(TAG_REGEX, s):
        s = re.sub(TAG_REGEX, _replace_tag, s)
    return s


def autoload(commander, modules: Sequence[str]):
    compositions = []
    for module_name in modules:
        module = importlib.import_module(module_name)
        compositions.append(module.command)
    commander.compose(*compositions)


def noop(*args, **kwargs):
    """ a function that does nothing except accepting everything you throw at it """
    pass


def _undefined_result(obj_key):
    return ValueError('unsupported path key "{}"'.format(obj_key))


def read_object_path(obj, obj_path, default=_undefined_result):
    def _default(exc=None):
        # raises the passed exception if no default value was provided
        # or returns the default value or None in case there’s no
        # exception
        if exc:
            if default is _undefined_result:
                raise exc
            else:
                return default
        else:
            return None if default is _undefined_result else default

    # if the object is None to begin with we just return the default
    if obj is None:
        return _default()

    # try to split one path token from the path and in case that
    # fails assume that the object_path is an attribute name
    try:
        obj_key, new_obj_path = obj_path.split('.', 1)
    except ValueError:
        obj_key, new_obj_path = obj_path, None
    result = _undefined_result

    # try to match one of the key regular expressions
    for matcher in (PATH_ATTRIBUTE, PATH_DICT, PATH_INDEX, PATH_CALL):
        match = re.match(matcher, obj_key)

        # continue with the next regex if we got no match
        if not match:
            continue

        # as all property keys refer to an attribute first
        # lets extract that attribute right now
        try:
            attr = getattr(obj, match.group('attr'))
        except AttributeError as exc:
            return _default(exc)

        if matcher == PATH_ATTRIBUTE:
            # we matched an attribute and already got that
            result = attr
            break
        elif matcher == PATH_DICT:
            # object key referred to a dictionary
            try:
                result = attr[match.group('key')]
            except KeyError as exc:
                return _default(exc)
            break
        elif matcher == PATH_INDEX:
            # object key referred to a sequence
            try:
                result = attr[match.group('index')]
            except IndexError as exc:
                return _default(exc)
            break
        elif matcher == PATH_CALL:
            # object key referred to a function call
            try:
                result = attr()
            except TypeError as exc:
                return _default(exc)
            break

    # none of the matchers matched, so raise an exception
    # if no default has been provided
    if result == _undefined_result:
        return _default(_undefined_result(obj_key))

    # there’s no new object key to process. that’s it!
    if new_obj_path is None:
        return result
    # show must go on. we read the next object key
    # from the result of this run
    else:
        return read_object_path(result, new_obj_path)


def failsafe(exc_type=Exception, enable=True, callback=None):
    """ keep a function from failing
    :param exc_type: the exception type that should be catched
    :param enable: if exception handling should be enabled
    :param callback: an optional callback for processing and on-failure return value
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            if enable:
                try:
                    return func(*args, **kwargs)
                except exc_type as exc:
                    if callback:
                        return callback(exc, args, kwargs)
            else:
                return func(*args, **kwargs)
        return wrapper
    return decorator


class FragileStreamHandler(logging.StreamHandler):
    """ simple wrapper around StreamHandler that handles broken output targets gracefully

    This is relevant for interfaces, that may disconnect at any time (e.g. a USB gadget interface).
    Here we need to ignore logging errors.
    """

    def handleError(self, record):
        # silently ignore logging errors
        pass



from kien import create_commander, var, CommandResult
from kien.command import help, quit
from kien.runner import ConsoleRunner

command = create_commander('destiny’s child')


@command('say', 'my', 'name', var('name'))
def say_my_name(name):
    message = 'baby, i mean {}, i love you'.format(name)
    yield CommandResult(True, message, dict(name=name, message=message))


class MyRunner(ConsoleRunner):
    def configure(self) -> None:
        super().configure()
        root = create_commander('root')
        root.compose(command, help.command, quit.command)
        self.commander = root


if __name__ == '__main__':
    try:
        runner = MyRunner()
        runner.run()
    except KeyboardInterrupt:
        pass



from functools import update_wrapper, wraps
import re
from typing import Any


def validate(**fields):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for field, validator in fields.items():
                try:
                    field_value = kwargs[field]
                except KeyError:
                    continue
                try:
                    validate_value(validator, field_value)
                except ValidationError as exc:
                    exc.field = field
                    raise exc
            return func(*args, **kwargs)
        return wrapper
    return decorator


def validate_value(validator, value):
    if getattr(validator, '__is_validator', False):
        # syntactic sugar for uses of uninstantiated validators
        validator = validator()
    validator.validate(value)


class ValidationError(ValueError):
    def __init__(self, *args: Any, field=None) -> None:
        super().__init__(*args)
        self.field = field


class Validatable:
    @classmethod
    def validate(cls, value):
        raise NotImplementedError()


class _AbstractValidator(Validatable):
    def __call__(self, value):
        self.validate(value)

    def __or__(self, other):
        return _Or(self, other)

    def __and__(self, other):
        return _And(self, other)


class _MultipleValidator(_AbstractValidator):
    def __init__(self, validator) -> None:
        self.validator = validator

    def validate(self, values):
        for value in values:
            validate_value(self.validator, value)


class Validator(_AbstractValidator):
    def __init__(self, func, args, kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def validate(self, value):
        self.func(*self.args, value, **self.kwargs)


class _Or(_AbstractValidator):
    def __init__(self, first, second):
        self.first = first
        self.second = second

    def validate(self, value):
        exceptions = []

        def try_value(validator):
            try:
                validate_value(validator, value)
                return True
            except ValidationError as exc:
                exceptions.append(exc)
                return False

        if try_value(self.first):
            return
        elif try_value(self.second):
            return
        else:
            raise ValidationError(' or '.join([str(exc) for exc in exceptions]))


class _And(_AbstractValidator):
    def __init__(self, first, second):
        self.first = first
        self.second = second

    def validate(self, value):
        validate_value(self.first, value)
        validate_value(self.second, value)


def simple_validator(func):
    def decorator(*args, **kwargs):
        validator = Validator(func, args, kwargs)
        update_wrapper(validator, func)
        return validator
    decorator.__is_validator = True
    return decorator


def list_of(validator):
    return _MultipleValidator(validator)


@simple_validator
def is_int(exact, value=None):
    if value is None:
        exact, value = None, exact
    try:
        value = int(value)
    except ValueError as exc:
        raise ValidationError('must be an integer') from exc

    if exact is not None and exact != value:
        raise ValidationError('must be exactly %d' % exact)


@simple_validator
def is_float(exact, value=None):
    if value is None:
        exact, value = None, exact
    try:
        value = float(value)
    except ValueError as exc:
        raise ValidationError('must be a float') from exc

    if exact is not None and exact != value:
        raise ValidationError('must be exactly %d' % exact)


@simple_validator
def is_gte(gte_value, value):
    if not float(value) >= gte_value:
        raise ValidationError('must be greater than or equal to %d' % gte_value)


@simple_validator
def is_gt(gt_value, value):
    if not float(value) > gt_value:
        raise ValidationError('must be greater than %d' % gt_value)


@simple_validator
def is_lte(lte_value, value):
    if not float(value) <= lte_value:
        raise ValidationError('must be less than or equal to %d' % lte_value)


@simple_validator
def is_lt(lt_value, value):
    if not float(value) < lt_value:
        raise ValidationError('must be less than %d' % lt_value)


@simple_validator
def is_between(min, max, value):
    if not (min < float(value) < max):
        raise ValidationError('must be between %d and %d' % (min, max))


@simple_validator
def is_equal(eq_value, value):
    if not eq_value == value:
        raise ValidationError('must be equal to %s' % str(eq_value))


@simple_validator
def identity(id_value, value):
    if id_value is not value:
        raise ValidationError('must be same as %s' % str(id_value))


@simple_validator
def length(value, min=None, max=None, exact=None):
    value_length = len(value)
    if exact is not None and value_length != exact:
        raise ValidationError('length must be %d' % exact)
    if min is not None and value_length < min:
        raise ValidationError('length must be greater than %d' % min)
    if max is not None and value_length > max:
        raise ValidationError('length must be less than %d' % max)


@simple_validator
def one_of(choices, value):
    if value not in choices:
        raise ValidationError('must be one of: %s' % ', '.join(map(str, choices)))


@simple_validator
def regex(expr, value, message=None):
    if re.fullmatch(expr, value) is None:
        raise ValidationError(message if message is not None else 'has invalid format')
